package org.eclipse.jdt.internal.compiler.apt.dispatch;

import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import java.lang.annotation.Annotation;
import java.util.Iterator;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import javax.lang.model.util.ElementFilter;
import java.util.HashSet;
import org.eclipse.jdt.internal.compiler.apt.model.TypeElementImpl;
import javax.lang.model.element.ElementKind;
import java.util.Collections;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import java.util.Set;
import org.eclipse.jdt.internal.compiler.apt.model.Factory;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import org.eclipse.jdt.internal.compiler.apt.util.ManyToMany;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import javax.annotation.processing.RoundEnvironment;

public class RoundEnvImpl implements RoundEnvironment
{
    private final BaseProcessingEnvImpl _processingEnv;
    private final boolean _isLastRound;
    private final CompilationUnitDeclaration[] _units;
    private final ManyToMany<TypeElement, Element> _annoToUnit;
    private final ReferenceBinding[] _binaryTypes;
    private final Factory _factory;
    private Set<Element> _rootElements;
    
    public RoundEnvImpl(CompilationUnitDeclaration[] units, ReferenceBinding[] binaryTypeBindings, boolean isLastRound, BaseProcessingEnvImpl env) {
        this._rootElements = null;
        this._processingEnv = env;
        this._isLastRound = isLastRound;
        this._units = units;
        this._factory = this._processingEnv.getFactory();
        AnnotationDiscoveryVisitor visitor = new AnnotationDiscoveryVisitor(this._processingEnv);
        if (this._units != null) {
            CompilationUnitDeclaration[] units2;
            CompilationUnitDeclaration unit;
            for (int length = (units2 = this._units).length, i = 0; i < length; unit.scope.suppressImportErrors = false, ++i) {
                unit = units2[i];
                unit.scope.suppressImportErrors = true;
                unit.traverse(visitor, unit.scope);
            }
        }
        this._annoToUnit = visitor._annoToElement;
        if (binaryTypeBindings != null) {
            this.collectAnnotations(binaryTypeBindings);
        }
        this._binaryTypes = binaryTypeBindings;
    }
    
    private void collectAnnotations(ReferenceBinding[] referenceBindings) {
        for (ReferenceBinding referenceBinding : referenceBindings) {
            if ((referenceBinding instanceof ParameterizedTypeBinding)) {
                referenceBinding = ((ParameterizedTypeBinding)referenceBinding).genericType();
            }
            for (AnnotationBinding annotationBinding : Factory.getPackedAnnotationBindings(referenceBinding.getAnnotations())) {
                TypeElement anno = (TypeElement)this._factory.newElement(annotationBinding.getAnnotationType());
                Element element = this._factory.newElement(referenceBinding);
                this._annoToUnit.put(anno, element);
            }
            for (FieldBinding fieldBinding : referenceBinding.fields()) {
                for (AnnotationBinding annotationBinding2 : Factory.getPackedAnnotationBindings(fieldBinding.getAnnotations())) {
                    TypeElement anno2 = (TypeElement)this._factory.newElement(annotationBinding2.getAnnotationType());
                    Element element2 = this._factory.newElement(fieldBinding);
                    this._annoToUnit.put(anno2, element2);
                }
            }
            for (MethodBinding methodBinding : referenceBinding.methods()) {
                for (AnnotationBinding annotationBinding3 : Factory.getPackedAnnotationBindings(methodBinding.getAnnotations())) {
                    TypeElement anno3 = (TypeElement)this._factory.newElement(annotationBinding3.getAnnotationType());
                    Element element3 = this._factory.newElement(methodBinding);
                    this._annoToUnit.put(anno3, element3);
                }
            }
            ReferenceBinding[] memberTypes = referenceBinding.memberTypes();
            this.collectAnnotations(memberTypes);
        }
    }
    
    public Set<TypeElement> getRootAnnotations() {
        return Collections.unmodifiableSet(this._annoToUnit.getKeySet());
    }
    
    @Override
    public boolean errorRaised() {
        return this._processingEnv.errorRaised();
    }
    
    @Override
    public Set<? extends Element> getElementsAnnotatedWith(TypeElement a) {
        if (a.getKind() != ElementKind.ANNOTATION_TYPE) {
            throw new IllegalArgumentException("Argument must represent an annotation type");
        }
        Binding annoBinding = ((TypeElementImpl)a)._binding;
        if (0x0L != (annoBinding.getAnnotationTagBits() & 0x1000000000000L)) {
            Set<Element> annotatedElements = new HashSet(this._annoToUnit.getValues(a));
            ReferenceBinding annoTypeBinding = (ReferenceBinding)annoBinding;
            for (TypeElement element : ElementFilter.typesIn(this.getRootElements())) {
                ReferenceBinding typeBinding = (ReferenceBinding)((TypeElementImpl)element)._binding;
                this.addAnnotatedElements(annoTypeBinding, typeBinding, annotatedElements);
            }
            return Collections.unmodifiableSet(annotatedElements);
        }
        return Collections.unmodifiableSet(this._annoToUnit.getValues(a));
    }
    
    private void addAnnotatedElements(ReferenceBinding anno, ReferenceBinding type, Set<Element> result) {
        if (type.isClass() && this.inheritsAnno(type, anno)) {
            result.add(this._factory.newElement(type));
        }
        ReferenceBinding[] memberTypes;
        for (int length = (memberTypes = type.memberTypes()).length, i = 0; i < length; ++i) {
            ReferenceBinding element = memberTypes[i];
            this.addAnnotatedElements(anno, element, result);
        }
    }
    
    private boolean inheritsAnno(ReferenceBinding element, ReferenceBinding anno) {
        ReferenceBinding searchedElement = element;
        do {
            if ((searchedElement instanceof ParameterizedTypeBinding)) {
                searchedElement = ((ParameterizedTypeBinding)searchedElement).genericType();
            }
            AnnotationBinding[] annos = Factory.getPackedAnnotationBindings(searchedElement.getAnnotations());
            int length = annos.length;
            int i = 0;
            while (i < length) {
                AnnotationBinding annoBinding = annos[i];
                if (annoBinding.getAnnotationType() == anno) {
                    return true;
                }
                ++i;
            }
        } while ((searchedElement = searchedElement.superclass()) != null);
        return false;
    }
    
    @Override
    public Set<? extends Element> getElementsAnnotatedWith(Class<? extends Annotation> a) {
        String canonicalName = a.getCanonicalName();
        if (canonicalName == null) {
            throw new IllegalArgumentException("Argument must represent an annotation type");
        }
        TypeElement annoType = this._processingEnv.getElementUtils().getTypeElement(canonicalName);
        if (annoType == null) {
            return Collections.emptySet();
        }
        return this.getElementsAnnotatedWith(annoType);
    }
    
    @Override
    public Set<? extends Element> getRootElements() {
        if (this._units == null) {
            return Collections.emptySet();
        }
        if (this._rootElements == null) {
            Set<Element> elements = new HashSet(this._units.length);
            CompilationUnitDeclaration[] units;
            for (int length = (units = this._units).length, i = 0; i < length; ++i) {
                CompilationUnitDeclaration unit = units[i];
                if (unit.scope != null && unit.scope.topLevelTypes != null) {
                    SourceTypeBinding[] topLevelTypes;
                    int length2 = (topLevelTypes = unit.scope.topLevelTypes).length;
                    int j = 0;
                    while (j < length2) {
                        SourceTypeBinding binding = topLevelTypes[j];
                        Element element = this._factory.newElement(binding);
                        if (element == null) {
                            throw new IllegalArgumentException("Top-level type binding could not be converted to element: " + binding);
                        }
                        elements.add(element);
                        ++j;
                    }
                }
            }
            if (this._binaryTypes != null) {
                ReferenceBinding[] binaryTypes;
                int length3 = (binaryTypes = this._binaryTypes).length;
                int k = 0;
                while (k < length3) {
                    ReferenceBinding typeBinding = binaryTypes[k];
                    Element element2 = this._factory.newElement(typeBinding);
                    if (element2 == null) {
                        throw new IllegalArgumentException("Top-level type binding could not be converted to element: " + typeBinding);
                    }
                    elements.add(element2);
                    ++k;
                }
            }
            this._rootElements = elements;
        }
        return this._rootElements;
    }
    
    @Override
    public boolean processingOver() {
        return this._isLastRound;
    }
}
